from openfisca_core.model_api import *
from openfisca_us.entities import *
from openfisca_us.tools.general import *
from openfisca_us.variables.demographic.tax_unit import MARSType


class dsi(Variable):
    value_type = bool
    entity = TaxUnit
    definition_period = YEAR
    documentation = (
        """True if claimed as dependent on another return; otherwise false"""
    )


class eic(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Number of EIC qualifying children (range: 0 to 3)"""

    def formula(tax_unit, period, parameters):
        return tax_unit.max(
            tax_unit.members("is_tax_unit_head", period)
            * tax_unit.members("age", period)
        )


class flpdyr(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Calendar year for which taxes are calculated"""


class mars(Variable):
    value_type = Enum
    entity = TaxUnit
    possible_values = MARSType
    default_value = MARSType.SINGLE
    definition_period = YEAR
    label = "Marital status for the tax unit"

    def formula(tax_unit, period, parameters):
        has_spouse = tax_unit.max(
            tax_unit.members("is_tax_unit_spouse", period)
            * (tax_unit.members("age", period) > 0)
        )
        return where(has_spouse, MARSType.JOINT, MARSType.SINGLE)


class midr(Variable):
    value_type = bool
    entity = TaxUnit
    definition_period = YEAR
    documentation = (
        """True if separately filing spouse itemizes; otherwise false"""
    )


class recid(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Unique numeric identifier for filing unit; appears as RECID variable in tc CLI minimal output"""


class xtot(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Total number of exemptions for filing unit"""


class age_head(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Age in years of taxpayer (i.e. primary adult)"""

    def formula(tax_unit, period, parameters):
        return tax_unit.max(
            tax_unit.members("is_tax_unit_head", period)
            * tax_unit.members("age", period)
        )


class age_spouse(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = (
        """Age in years of spouse (i.e. secondary adult if present)"""
    )

    def formula(tax_unit, period, parameters):
        return tax_unit.max(
            tax_unit.members("is_tax_unit_spouse", period)
            * tax_unit.members("age", period)
        )


class agi_bin(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Historical AGI category used in data extrapolation"""


class blind_head(Variable):
    value_type = bool
    entity = TaxUnit
    definition_period = YEAR
    documentation = """True if taxpayer is blind; otherwise False"""


class blind_spouse(Variable):
    value_type = bool
    entity = TaxUnit
    definition_period = YEAR
    documentation = """1 if spouse is blind; otherwise 0"""


class cmbtp(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Estimate of income on (AMT) Form 6251 but not in AGI"""


class filer_cmbtp(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Income on Form 6251 not in AGI for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("cmbtp", tax_unit, period)


class e00200(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = (
        """Wages, salaries, and tips net of pension contributions"""
    )


class filer_e00200(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Wages, salaries, and tips for filing unit (excluding dependents) net of pension contributions (pencon)"""

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e00200", tax_unit, period)


class pencon(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Contributions to defined-contribution pension plans"""


class filer_pencon(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Contributions to defined-contribution pension plans for filing unit (excluding dependents)"""

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("pencon", tax_unit, period)


class e00300(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Taxable interest income"""


class filer_e00300(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Taxable interest income for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e00300", tax_unit, period)


class e00400(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Tax-exempt interest income"""


class filer_e00400(Variable):
    value_type = float
    entity = TaxUnit
    label = (
        u"Tax-exempt interest income for the tax unit (excluding dependents)"
    )
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e00400", tax_unit, period)


class e00600(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Ordinary dividends included in AGI"""


class filer_e00600(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Ordinary dividends for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e00600", tax_unit, period)


class e00650(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Qualified dividends included in ordinary dividends"""


class filer_e00650(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Qualified dividends included in ordinary dividends for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e00650", tax_unit, period)


class e00700(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Taxable refunds of state and local income taxes"""


class filer_e00700(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Taxable refunds of state/local taxes for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e00700", tax_unit, period)


class e00800(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Alimony received"""


class filer_e00800(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Alimony for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e00800", tax_unit, period)


class e00900(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Sch C business net profit/loss"""


class filer_e00900(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Sch C business net profit/loss for filing unit (excluding dependents)"""

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e00900", tax_unit, period)


class e01100(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Capital gain distributions not reported on Sch D"""


class filer_e01100(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Capital gains not reported on Sch D for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e01100", tax_unit, period)


class e01200(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Other net gain/loss from Form 4797"""


class filer_e01200(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Other net gain/loss for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e01200", tax_unit, period)


class e01400(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Taxable IRA distributions"""


class filer_e01400(Variable):
    value_type = float
    entity = TaxUnit
    label = (
        u"Taxable IRA distributions for the tax unit (excluding dependents)"
    )
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e01400", tax_unit, period)


class e01500(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Total pensions and annuities"""


class filer_e01500(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Pensions and annuities for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e01500", tax_unit, period)


class e01700(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Taxable pensions and annuities"""


class filer_e01700(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Taxable pensions and annuities for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e01700", tax_unit, period)


class e02000(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Sch E total rental, royalty, partnership, S-corporation, etc, income/loss (includes e26270 and e27200)"""


class filer_e02000(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Rentals, royalties etc. for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e02000", tax_unit, period)


class e02100(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Farm net income/loss from Sch F"""


class filer_e02100(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Farm net income/loss for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e02100", tax_unit, period)


class e02300(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Unemployment insurance benefits"""


class filer_e02300(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Unemployment insurance benefits for filing unit (excluding dependents)"""

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e02300", tax_unit, period)


class e02400(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Total social security (OASDI) benefits"""


class filer_e02400(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Social security benefits for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e02400", tax_unit, period)


class e03150(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Total deductible IRA contributions"""


class filer_e03150(Variable):
    value_type = float
    entity = TaxUnit
    label = (
        u"Deductible IRA contributions for the tax unit (excluding dependents)"
    )
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e03150", tax_unit, period)


class e03210(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Student loan interest"""


class filer_e03210(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Student loan interest for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e03210", tax_unit, period)


class e03220(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Educator expenses"""


class filer_e03220(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Educator expenses for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e03220", tax_unit, period)


class e03230(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Tuition and fees from Form 8917"""


class filer_e03230(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Tuition for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e03230", tax_unit, period)


class e03240(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Domestic production activities from Form 8903"""


class filer_e03240(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Domestic production activities for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e03240", tax_unit, period)


class e03270(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Self-employed health insurance deduction"""


class filer_e03270(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Self-employed health insurance for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e03270", tax_unit, period)


class e03290(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Health savings account deduction from Form 8889"""


class filer_e03290(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Health savings account deduction for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e03290", tax_unit, period)


class e03300(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Contributions to SEP, SIMPLE and qualified plans"""


class filer_e03300(Variable):
    value_type = float
    entity = TaxUnit
    label = u"SEP, SIMPLE, etc. contributions for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e03300", tax_unit, period)


class e03400(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Penalty on early withdrawal of savings"""


class filer_e03400(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Early savings withdrawal penalty for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e03400", tax_unit, period)


class e03500(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Alimony paid"""


class filer_e03500(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Alimony paid for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e03500", tax_unit, period)


class e07240(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = (
        """Retirement savings contributions credit from Form 8880"""
    )


class e07260(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Residential energy credit from Form 5695"""


class e07300(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Foreign tax credit from Form 1116"""


class filer_e07300(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Foreign tax credit (Form 1114) for the tax unit (excluding dependents)"
    unit = "currency-USD"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e07300", tax_unit, period)


class e07400(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """General business credit from Form 3800"""


class e07600(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Prior year minimum tax credit from Form 8801"""


class e09700(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Recapture of Investment Credit"""


class e09800(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Unreported payroll taxes from Form 4137 or 8919"""


class e09900(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Penalty tax on qualified retirement plans"""


class e11200(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Excess payroll (FICA/RRTA) tax withheld"""


class e17500(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Itemizable medical and dental expenses.  WARNING: this variable is zero below the floor in PUF data."""


class filer_e17500(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Itemized medical and dental expenses for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e17500", tax_unit, period)


class e18400(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Itemizable state and local income/sales taxes"""


class filer_e18400(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Itemized SALT for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e18400", tax_unit, period)


class e18500(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Itemizable real-estate taxes paid"""


class filer_e18500(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Itemized real estate for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e18500", tax_unit, period)


class e19200(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Itemizable interest paid"""


class filer_e19200(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Itemizable interest paid for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e19200", tax_unit, period)


class e19800(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Itemizable charitable giving: cash/check contributions.  WARNING: this variable is already capped in PUF data."""


class filer_e19800(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Itemized charity for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e19800", tax_unit, period)


class e20100(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Itemizable charitable giving: other than cash/check contributions.  WARNING: this variable is already capped in PUF data."""


class filer_e20100(Variable):
    value_type = float
    entity = TaxUnit
    label = (
        u"Itemized non-cash charity for the tax unit (excluding dependents)"
    )
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e20100", tax_unit, period)


class e20400(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Itemizable miscellaneous deductions.  WARNING: this variable is zero below the floor in PUF data."""


class filer_e20400(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Itemized misc. for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e20400", tax_unit, period)


class g20500(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Itemizable gross (before 10% AGI disregard) casualty or theft loss"""


class filer_g20500(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Itemized casualty loss for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("g20500", tax_unit, period)


class e24515(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Sch D: Un-Recaptured Section 1250 Gain"""


class filer_e24515(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Section 1250 Gain (un-recaptured) for the tax unit (excluding dependents)"
    unit = "currency-USD"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e24515", tax_unit, period)


class e24518(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Sch D: 28% Rate Gain or Loss"""


class filer_e24518(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Sch D Rate Gain or Loss for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e24518", tax_unit, period)


class e26270(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Sch E: Combined partnership and S-corporation net income/loss (includes k1bx14p and k1bx14s amounts and is included in e02000)"""


class filer_e26270(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Label for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e26270", tax_unit, period)


class e27200(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = (
        """Sch E: Farm rent net income or loss (included in e02000)"""
    )


class filer_e27200(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Label for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e27200", tax_unit, period)


class e32800(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Child/dependent-care expenses for qualifying persons from Form 2441"""


class filer_e32800(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Child/dependent-care expenses for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e32800", tax_unit, period)


class e58990(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Investment income elected amount from Form 4952"""


class filer_e58990(Variable):
    value_type = float
    entity = TaxUnit
    label = u"Investment income (Form 3952) for the tax unit (excluding dependents)"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e58990", tax_unit, period)


class e62900(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = (
        """Alternative Minimum Tax foreign tax credit from Form 6251"""
    )


class filer_e62900(Variable):
    value_type = float
    entity = TaxUnit
    label = u"AMT foreign tax credit (Form 6251) for the tax unit (excluding dependents)"
    unit = "currency-USD"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("e62900", tax_unit, period)


class e87530(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    label = "Qualified tuition expenses"
    documentation = (
        "Adjusted qualified lifetime learning expenses for all students"
    )


qualified_tuition_expenses = variable_alias(
    "qualified_tuition_expenses", e87530
)


class elderly_dependents(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = """number of dependents age 65+ in filing unit excluding taxpayer and spouse"""


class f2441(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = """number of child/dependent-care qualifying persons"""


class f6251(Variable):
    value_type = bool
    entity = TaxUnit
    definition_period = YEAR
    documentation = (
        """True if Form 6251 (AMT) attached to return; otherwise false"""
    )


class a_lineno(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = """CPS line number for the person record of the head of the tax filing unit (not used in tax-calculation logic)"""


class ffpos(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = """CPS family identifier within household (not used in tax-calculation logic)"""


class fips(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = """FIPS state code (not used in tax-calculation logic)"""


class h_seq(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = (
        """CPS household sequence number (not used in tax-calculation logic)"""
    )


class data_source(Variable):
    value_type = bool
    entity = TaxUnit
    definition_period = YEAR
    documentation = """True if unit is created primarily from IRS-SOI PUF data; false if created primarily from CPS data (not used in tax-calculation logic)"""


class k1bx14(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = (
        """Partner self-employment earnings/loss (included in e26270 total)"""
    )


class filer_k1bx14(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Partner self-employment earnings/loss for tax unit (excluding dependents) (included in e26270 total)"""

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("k1bx14", tax_unit, period)


class mcaid_ben(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Imputed Medicaid benefits expressed as the actuarial value of Medicaid health insurance"""


class mcare_ben(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Imputed Medicare benefits expressed as the actuarial value of Medicare health insurance"""


class n24(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Number of children who are Child-Tax-Credit eligible, one condition for which is being under age 17"""


class nu06(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Number of dependents under 6 years old"""


class nu13(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Number of dependents under 13 years old"""


class nu18(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = (
        """Number of people under 18 years old in the filing unit"""
    )


class n1820(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = (
        """Number of people age 18-20 years old in the filing unit"""
    )


class n21(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    documentation = (
        """Number of people 21 years old or older in the filing unit"""
    )


class other_ben(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Non-imputed benefits"""


class p08000(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Other tax credits (but not including Sch R credit)"""


class p22250(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Sch D: Net short-term capital gains/losses"""


class filer_p22250(Variable):
    value_type = float
    entity = TaxUnit
    label = (
        u"Net short-term capital gains for the tax unit (excluding dependents)"
    )
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("p22250", tax_unit, period)


class p23250(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = """Sch D: Net long-term capital gains/losses"""


class filer_p23250(Variable):
    value_type = float
    entity = TaxUnit
    label = (
        u"Net long-term capital gains for the tax unit (excluding dependents)"
    )
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("p23250", tax_unit, period)


class e87521(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Total tentative AmOppCredit amount for all students"""


class s006(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Filing unit sampling weight; appears as WEIGHT variable in tc CLI minimal output"""


class snap_ben(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Imputed SNAP benefits"""


class housing_ben(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Imputed housing benefits"""


class ssi_ben(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Imputed SSI benefits"""


class tanf_ben(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Imputed TANF benefits"""


class vet_ben(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Imputed Veteran's benefits"""


class wic_ben(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Imputed WIC benefits"""


class pt_sstb_income(Variable):
    value_type = bool
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Whether business income is from a specified service trade or business (SSTB), rather than from a qualified trade or business"""


class pt_binc_w2_wages(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Filing unit's share of total W-2 wages paid by the pass-through business"""


class pt_ubia_property(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = """Filing unit's share of total business property owned by the pass-through business"""



from openfisca_core.model_api import *
from openfisca_us.entities import *
from openfisca_us.tools.general import *


class ssdi(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = "Social Security Disability Insurance amount"
    label = "Social Security Disability Insurance"



from openfisca_core.model_api import *
from openfisca_us.entities import *
from openfisca_us.tools.general import *


class spm_unit_total_income(Variable):
    value_type = float
    entity = SPMUnit
    label = u"SPM unit total income"
    definition_period = YEAR
    unit = "currency-USD"


class spm_unit_snap(Variable):
    value_type = float
    entity = SPMUnit
    label = u"SPM unit SNAP subsidy"
    definition_period = YEAR
    unit = "currency-USD"


class spm_unit_capped_housing_subsidy(Variable):
    value_type = float
    entity = SPMUnit
    label = u"SPM unit capped housing subsidy"
    definition_period = YEAR
    unit = "currency-USD"


class spm_unit_school_lunch_subsidy(Variable):
    value_type = float
    entity = SPMUnit
    label = u"SPM unit school lunch subsidy"
    definition_period = YEAR
    unit = "currency-USD"


class spm_unit_energy_subsidy(Variable):
    value_type = float
    entity = SPMUnit
    label = u"SPM unit school energy subsidy"
    definition_period = YEAR
    unit = "currency-USD"


class spm_unit_wic(Variable):
    value_type = float
    entity = SPMUnit
    label = u"SPM unit WIC subsidy"
    definition_period = YEAR
    unit = "currency-USD"


class spm_unit_net_income(Variable):
    value_type = float
    entity = SPMUnit
    label = u"SPM unit net income"
    definition_period = YEAR
    unit = "currency-USD"

    def formula(spm_unit, period, parameters):
        INCOME_COMPONENTS = [
            "spm_unit_total_income",
            "spm_unit_snap",
            "spm_unit_capped_housing_subsidy",
            "spm_unit_school_lunch_subsidy",
            "spm_unit_energy_subsidy",
            "spm_unit_wic",
        ]
        EXPENSE_COMPONENTS = [
            "spm_unit_fica",
            "spm_unit_federal_tax",
            "spm_unit_state_tax",
            "spm_unit_capped_work_childcare_expenses",
            "spm_unit_medical_expenses",
        ]
        income = add(spm_unit, period, *INCOME_COMPONENTS)
        expense = add(spm_unit, period, *EXPENSE_COMPONENTS)
        return income - expense


class spm_unit_spm_threshold(Variable):
    value_type = float
    entity = SPMUnit
    label = u"SPM unit's SPM poverty threshold"
    definition_period = YEAR
    unit = "currency-USD"


class spm_unit_is_in_spm_poverty(Variable):
    value_type = bool
    entity = SPMUnit
    label = u"SPM unit in SPM poverty"
    definition_period = YEAR

    def formula(spm_unit, period, parameters):
        income = spm_unit("spm_unit_net_income", period)
        poverty_threshold = spm_unit("spm_unit_spm_threshold", period)
        return income < poverty_threshold


class experienced_covid_income_loss(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Experienced Covid income loss"
    documentation = "Whether the SPM unit experienced a loss of income due to COVID-19 since February 2020"
    definition_period = YEAR



from openfisca_us.model_api import *


class fcc_fpg_ratio(Variable):
    value_type = float
    entity = SPMUnit
    label = "Federal poverty ratio per FCC"
    documentation = "SPM unit's ratio of IRS gross income to their federal poverty guideline"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/cfr/text/47/54.400#f"
    unit = "ratio"

    def formula(spm_unit, period, parameters):
        income = spm_unit.sum(spm_unit.members("irs_gross_income", period))
        fpg = spm_unit("spm_unit_fpg", period)
        return income / fpg



from openfisca_us.model_api import *


class meets_snap_categorical_eligibility(Variable):
    value_type = bool
    entity = SPMUnit
    label = "SNAP categorical eligibility"
    documentation = "Whether this SPM unit is eligible for SNAP benefits via participation in other programs"
    definition_period = YEAR
    reference = "https://fns-prod.azureedge.net/sites/default/files/resource-files/fna-2008-amended-through-pl-116-94.pdf#page=11"

    def formula(spm_unit, period, parameters):
        programs = parameters(period).usda.snap.categorical_eligibility
        return np.any([spm_unit(program, period) for program in programs])



from openfisca_us.model_api import *


class is_ssi_disabled(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    documentation = "Indicates whether a person is disabled for the Supplemental Security Income program"
    label = "SSI disabled"



from openfisca_us.model_api import *


class is_disabled_or_elderly_for_snap(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    label = "Is disabled or elderly for SNAP"
    documentation = "Indicates that a person is defined as disabled or elderly based on the USDA definition"



from openfisca_us.model_api import *


class SNAPUttilityAllowanceType(Enum):
    SUA = "Standard Utility Allowance"
    LUA = "Limited Utility Allowance"
    TUA = "Telephone Utility Allowance"
    NONE = "None"


class snap_utility_allowance_type(Variable):
    value_type = Enum
    possible_values = SNAPUttilityAllowanceType
    entity = SPMUnit
    label = "SNAP utility allowance eligibility"
    default_value = SNAPUttilityAllowanceType.NONE
    documentation = (
        "The type of utility allowance that is eligible for the SPM unit"
    )
    definition_period = YEAR

    def formula(spm_unit, period, parameters):
        household = spm_unit.household
        return select(
            [
                household("has_heating_cooling_expense", period),
                household("has_other_utility_expense", period),
                household("has_telephone_expense", period),
                True,
            ],
            [
                SNAPUttilityAllowanceType.SUA,
                SNAPUttilityAllowanceType.LUA,
                SNAPUttilityAllowanceType.TUA,
                SNAPUttilityAllowanceType.NONE,
            ],
        )



from openfisca_tools import Microsimulation as GeneralMicrosimulation
from openfisca_us import CountryTaxBenefitSystem
from openfisca_us.entities import entities
from openfisca_us_data import CPS


class Microsimulation(GeneralMicrosimulation):
    tax_benefit_system = CountryTaxBenefitSystem
    entities = entities
    default_dataset = CPS



from openfisca_us.model_api import *


class is_snap_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "SNAP eligible"
    documentation = "Whether this SPM unit is eligible for SNAP benefits"
    definition_period = YEAR
    reference = (
        "https://www.law.cornell.edu/uscode/text/7/2017#a",
        "https://www.law.cornell.edu/uscode/text/7/2014#c",
    )

    def formula(spm_unit, period, parameters):
        meets_net_income_test = spm_unit("meets_snap_net_income_test", period)
        meets_gross_income_test = spm_unit(
            "meets_snap_gross_income_test", period
        )
        meets_asset_test = spm_unit("meets_snap_asset_test", period)
        normal_eligibility = (
            meets_net_income_test & meets_asset_test & meets_gross_income_test
        )
        # Categorical eligibility (SSI, TANF, and BBCE TANF) overrides tests.
        categorical_eligibility = spm_unit(
            "meets_snap_categorical_eligibility", period
        )
        return normal_eligibility | categorical_eligibility



from openfisca_us.model_api import *


class snap_assets(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = YEAR
    documentation = "Countable assets for SNAP limits"
    label = "SNAP assets"
    unit = "currency-USD"



from openfisca_us.model_api import *


class has_other_utility_expense(Variable):
    value_type = bool
    entity = Household
    label = "Has other utility expenses"
    documentation = "Whether the household has utility bills other than heating/cooling and telephone"
    definition_period = YEAR



from openfisca_us.model_api import *


class american_opportunity_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "American Opportunity Credit"
    unit = "currency-USD"
    documentation = "Total value of the American Opportunity Credit"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/25A#b"

    def formula(tax_unit, period, parameters):
        education = parameters(period).irs.credits.education
        aoc = education.american_opportunity_credit
        person = tax_unit.members
        is_eligible = person(
            "is_eligible_for_american_opportunity_credit", period
        )
        tuition_expenses = (
            person("qualified_tuition_expenses", period) * is_eligible
        )
        maximum_amount_per_student = aoc.amount.calc(tuition_expenses)
        maximum_amount = tax_unit.sum(maximum_amount_per_student)
        phaseout = tax_unit("education_credit_phaseout", period)
        return max_(0, maximum_amount * (1 - phaseout))


class refundable_american_opportunity_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Refundable American Opportunity Credit"
    unit = "currency-USD"
    documentation = (
        "Value of the refundable portion of the American Opportunity Credit"
    )
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/25A#i"

    def formula(tax_unit, period, parameters):
        aoc = parameters(
            period
        ).irs.credits.education.american_opportunity_credit
        return aoc.refundability * tax_unit(
            "american_opportunity_credit", period
        )


class non_refundable_american_opportunity_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Non-refundable American Opportunity Credit"
    unit = "currency-USD"
    documentation = "Value of the non-refundable portion of the American Opportunity Credit"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/25A#i"

    def formula(tax_unit, period):
        total = tax_unit("american_opportunity_credit", period)
        refundable = tax_unit("refundable_american_opportunity_credit", period)
        return total - refundable


class is_eligible_for_american_opportunity_credit(Variable):
    value_type = bool
    entity = Person
    label = "Eligible for American Opportunity Credit"
    documentation = "Whether the person is eligible for the AOC in respect of qualified tuition expenses for this tax year. The expenses must be for one of the first four years of post-secondary education, and the person must not have claimed the AOC for any four previous tax years."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/25A#b_2"

    def formula(person, period):
        # If the person's filing unit has a claim from Form 8863, use that to determine eligibility.
        return person.tax_unit("e87521", period) > 0


class education_credit_phaseout(Variable):
    value_type = float
    entity = TaxUnit
    label = "Education credit phase-out"
    unit = "/1"
    documentation = "Percentage of the American Opportunity and Lifetime Learning credits which are phased out"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        education = parameters(period).irs.credits.education
        agi = tax_unit("adjusted_gross_income", period)
        is_joint = tax_unit("tax_unit_is_joint", period)
        phaseout_start = where(
            is_joint,
            education.phaseout.start.joint,
            education.phaseout.start.single,
        )
        phaseout_length = where(
            is_joint,
            education.phaseout.length.joint,
            education.phaseout.length.single,
        )
        excess_agi = max(0, agi - phaseout_start)
        return min_(1, excess_agi / phaseout_length)



from openfisca_us.model_api import *


class meets_tanf_non_cash_income_test(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Meets income test for TANF non-cash benefit"
    documentation = "Income eligibility (gross income as a percent of the poverty line) for TANF non-cash benefit for SNAP BBCE"
    definition_period = YEAR

    def formula(spm_unit, period, parameters):
        state = spm_unit.household("state_code_str", period)
        limits = parameters(period).hhs.tanf.non_cash
        income_limit_fpg = limits.gross_income_limit_fpg[state]
        income_fpg = spm_unit("snap_gross_income_fpg_ratio", period)
        return income_fpg <= income_limit_fpg



from openfisca_us.model_api import *


class snap_standard_deduction(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = YEAR
    documentation = "Standard deduction for calculating SNAP benefit amount"
    label = "SNAP standard deduction"
    unit = "currency-USD"
    reference = "https://www.law.cornell.edu/uscode/text/7/2014#e_1"

    def formula(spm_unit, period, parameters):
        standard_deductions = parameters(period).usda.snap.standard_deduction
        state_group = spm_unit.household("state_group_str", period)
        capped_household_size = min_(spm_unit.nb_persons(), 6)
        return standard_deductions[state_group][capped_household_size] * 12



from openfisca_us.model_api import *


class snap_dependent_care_deduction(Variable):
    value_type = float
    entity = SPMUnit
    label = "SNAP dependent care deduction"
    unit = "currency-USD"
    documentation = "Deduction from SNAP gross income for dependent care"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/7/2014#e_3"



import json
import yaml
import argparse

TAXCALC_PARAM_FILE = (
    "/home/nikhil/pslmodels/Tax-Calculator/taxcalc/policy_current_law.json"
)

MARS_map = {
    "single": "SINGLE",
    "mjoint": "JOINT",
    "mseparate": "SEPARATE",
    "headhh": "HEAD_OF_HOUSEHOLD",
    "widow": "WIDOW",
}


def year_to_date(year: int) -> str:
    return f"{year}-01-01"


def convert_param(param: dict) -> dict:
    result = dict(description=param["title"], values={})
    for value in param["value"]:
        date = year_to_date(value["year"])
        val = float(value["value"])
        if "MARS" not in value:
            result["values"][date] = val
        else:
            mars = MARS_map[value["MARS"]]
            if mars not in result:
                result[mars] = dict(values={date: val})
            else:
                result[mars]["values"][date] = val
    result["metadata"] = dict(
        unit="currency-USD",
    )
    if result["values"] == {}:
        del result["values"]
    return yaml.dump(result, sort_keys=False).replace("'", "")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Utility to convert tax-calculator parameters to OpenFisca parameters."
    )
    parser.add_argument(
        "--tc-params",
        default=TAXCALC_PARAM_FILE,
        help="Path to the tax-calculator policy_current_law.json file.",
    )
    parser.add_argument(
        "parameter", help="The name of the parameter to convert"
    )
    args = parser.parse_args()

    with open(args.tc_params, "r") as f:
        tc = json.load(f)

    print(convert_param(tc[args.parameter]))



from openfisca_us.model_api import *


class snap_max_benefit(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = YEAR
    documentation = "Maximum benefit for SPM unit, based on the state group and household size."
    label = "SNAP maximum benefit"
    unit = "currency-USD"

    def formula(spm_unit, period, parameters):
        # TODO: Logic for families with >8 people
        snap_max_benefits = parameters(period).usda.snap.amount.main
        state_group = spm_unit.household("state_group_str", period)
        household_size = spm_unit.nb_persons()

        return snap_max_benefits[state_group][household_size] * 12



"""This file contains your country package's metadata and dependencies."""

from setuptools import find_packages, setup

setup(
    name="OpenFisca-US",
    version="0.17.0",
    author="Nikhil Woodruff",
    author_email="nikhil@policyengine.org",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering :: Information Analysis",
    ],
    description="OpenFisca tax and benefit system for the US",
    keywords="benefit microsimulation social tax",
    license="http://www.fsf.org/licensing/licenses/agpl-3.0.html",
    url="https://github.com/PolicyEngine/openfisca-us",
    include_package_data=True,  # Will read MANIFEST.in
    data_files=[
        (
            "share/openfisca/openfisca-country-template",
            ["CHANGELOG.md", "LICENSE", "README.md"],
        ),
    ],
    install_requires=[
        "OpenFisca-Core[web-api] >= 35.0.0",
        "openfisca_us_data >= 0.1.2",
        "microdf_python",
        "pandas",
        "tqdm",
        "requests",
        "OpenFisca-Tools>=0.1.9",
        "pyyaml",
    ],
    extras_require={
        "dev": [
            "autopep8 >= 1.5.4, < 2.0.0",
            "black",
            "wheel",
            "pytest",
            "setuptools",
            "jupyter-book",
            "coverage",
        ],
    },
    python_requires=">=3.7,<3.8",
    entry_points={},
    packages=find_packages(),
)



from openfisca_us.model_api import *


class snap_deductions(Variable):
    value_type = float
    entity = SPMUnit
    label = "SNAP income deductions"
    unit = "currency-USD"
    documentation = "Deductions made from gross income for SNAP benefits"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/7/2014#e"

    def formula(spm_unit, period, parameters):
        snap_deductions = parameters(period).usda.snap.deductions
        return add(spm_unit, period, *snap_deductions.allowed)



from openfisca_us.model_api import *


class snap_net_income_pre_shelter(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = YEAR
    documentation = "SNAP net income before the shelter deduction, needed as intermediate to calculate shelter deduction"
    label = "SNAP net income (pre-shelter)"
    unit = "currency-USD"
    reference = "https://www.law.cornell.edu/uscode/text/7/2014#e_6_A"

    def formula(spm_unit, period):
        income = spm_unit("snap_gross_income", period)
        DEDUCTIONS = ("snap_standard_deduction", "snap_earnings_deduction")
        deductions = add(spm_unit, period, *DEDUCTIONS)
        return max_(income - deductions, 0)



from time import time
from pathlib import Path
from typing import Callable, Dict, List, Tuple
from taxcalc import Calculator, Policy, Records
from openfisca_us.tools.dev.calcfunctions import *
import pandas as pd
import ast
import yaml
import argparse

np.random.seed(int(time()))


def extract_return(fname):
    for x in ast.walk(
        ast.parse(open(Path(__file__).parent / "calcfunctions.py").read())
    ):
        if not (isinstance(x, ast.FunctionDef)):
            continue
        if not (x.name == fname):
            continue
        for b in x.body:
            if isinstance(b, ast.Return):
                if isinstance(b.value, ast.Name):
                    yield b.value.id
                elif isinstance(b.value, ast.Tuple):
                    yield from [element.id for element in b.value.elts]


def get_inputs(fn: Callable) -> Tuple:
    """
    Return the inputs of a function.
    """
    return fn.__code__.co_varnames[: fn.__code__.co_argcount]


def get_outputs(fn: Callable) -> Tuple:
    """
    Return the outputs of a function.
    """
    return list(extract_return(fn.__name__))


policy = Policy()
TEST_YEAR = 2019
DEBUG_MODE = True
TAXCALC_CPS_CSV_FILE = "~/PSLmodels/Tax-Calculator/cps.csv"
cps = pd.read_csv(TAXCALC_CPS_CSV_FILE, nrows=1 if DEBUG_MODE else None)
policy.set_year(TEST_YEAR)


def generate_input_values(
    inputs: List[str], output_names: List[str], overrides: Dict[str, float]
) -> Dict[str, float]:
    """Fetches the parameter value if the input is a parameter,
    otherwise returns the value from a random CPS tax unit.

    Args:
        inputs (List[str]): The input names.
        output_names (List[str]): The names of the outputs.
        overrides (Dict[str, float]): A dictionary of overrides.

    Returns:
        Dict[str, float]: The generated values.
    """
    if not DEBUG_MODE:
        try:
            cps_tax_unit = cps[(cps[output_names] != 0).min(axis=1)].sample(
                n=1
            )
        except:
            cps_tax_unit = cps.sample(n=1)
    result = {}
    for input_ in inputs:
        try:
            result[input_] = getattr(policy, input_)[0]
        except AttributeError:
            if input_ in overrides:
                result[input_] = overrides[input_]
            elif DEBUG_MODE:
                result[input_] = 0
            else:
                result[input_] = int(cps_tax_unit[input_].values[0])
    return result


def generate_unit_test(
    fn_name: str, input_overrides: Dict[str, float] = None
) -> Tuple[dict, Callable[[dict], Tuple[float]], Tuple[float]]:
    """Generates a unit test for a calcfunctions function.

    Args:
        fn_name (str): The name of the taxcalc function.
        input_overrides (Dict[str, float]): A dictionary of overrides.

    Returns:
        Tuple[dict, Callable[[dict, Tuple[float]]], Tuple[float]]: Keyword args, the function taking them and the expected result.
    """
    if input_overrides is None:
        input_overrides = {}
    fn = CALCFUNCTIONS[fn_name]

    def fn_(inputs):
        result = fn(**inputs)
        if not isinstance(result, tuple):
            result = (result,)
        return result

    kwargs = generate_input_values(
        get_inputs(fn), get_outputs(fn), overrides=input_overrides
    )

    return kwargs, fn_, fn_(kwargs)


def attempt_generate_all_nonzero_unit_test(
    fn_name: str,
) -> Tuple[dict, Callable[[dict], Tuple[float]], Tuple[float]]:
    for _ in range(1_000):
        kwargs, fn_, result = generate_unit_test(fn_name)
        # This function often needs manually modifying depending on the function
        if all([i != 0 for i in result]):
            return kwargs, fn_, result


def generate_yaml_test_dict(fn_name: str, name: str = None):
    kwargs, _, result = attempt_generate_all_nonzero_unit_test(fn_name)
    outputs = get_outputs(CALCFUNCTIONS[fn_name])
    test_dict = dict(
        name=f"Unit test for {fn_name}" if name is None else name,
        period=2019,
        absolute_error_margin=1,
        input={
            rename_variable(x): translate_value(x, y)
            for x, y in kwargs.items()
            if x in cps and x not in outputs
        },
        output={
            rename_variable(name): float(res)
            for name, res in zip(outputs, result)
        },
    )
    return test_dict


def generate_yaml_tests(fn_name: str, n: int = 10) -> str:
    return yaml.dump(
        [
            generate_yaml_test_dict(
                fn_name,
                name=f"{fn_name} unit test {i + 1} (from generate_taxcalc_unit_tests.py)",
            )
            for i in range(n)
        ],
        sort_keys=False,
    )


RENAMES = {
    "MARS": "mars",
}

VALUE_RENAMES = {
    "MARS": {
        1: "SINGLE",
        2: "JOINT",
        3: "SEPARATE",
        4: "HOUSEHOLD_HEAD",
        5: "WIDOW",
    }
}

INVERSE_RENAMES = {y: x for x, y in RENAMES.items()}

INVERSE_VALUE_RENAMES = {
    x: {b: a for a, b in y.items()} for x, y in VALUE_RENAMES.items()
}


def rename_variable(x):
    return RENAMES.get(x, x)


def translate_value(name, x):
    return VALUE_RENAMES[name].get(x, x) if name in VALUE_RENAMES else x


def inverse_rename_variable(x):
    return INVERSE_RENAMES.get(x, x)


def inverse_translate_value(name, x):
    return (
        INVERSE_VALUE_RENAMES[name].get(x, x)
        if name in INVERSE_VALUE_RENAMES
        else x
    )


def debug_test_yaml(fn_name: str, file: str, i: int = 0):
    with open(file, "r") as f:
        tests = yaml.safe_load(f)
    test = tests[i]
    input_overrides = {
        inverse_rename_variable(x): inverse_translate_value(
            inverse_rename_variable(x), y
        )
        for x, y in test["input"].items()
    }
    kwargs, fn_, _ = generate_unit_test(fn_name, input_overrides)
    fn_(kwargs)


def generate_taxcalc_cps_df():
    r = Records.cps_constructor()
    p = Policy(year=TEST_YEAR)
    c = Calculator(policy=p, records=r)
    df = c.dataframe(None, all_vars=True)
    df.to_csv("cps.csv")


if __name__ == "__main__":
    if not DEBUG_MODE:
        parser = argparse.ArgumentParser(
            description="Generate unit tests for calcfunctions."
        )
        parser.add_argument(
            "function",
            type=str,
            help="The name of the function to generate unit tests for.",
        )
        parser.add_argument(
            "--n",
            type=int,
            default=10,
            help="The number of unit tests to generate.",
        )
        args = parser.parse_args()
        print(generate_yaml_tests(args.function, args.n))
    else:
        # An example of debugging a unit test
        debug_test_yaml(
            "GainsTax",
            "openfisca_us/tests/policy/baseline/calcfunctions/gainstax.yaml",
            2,
        )



from openfisca_core.taxbenefitsystems import TaxBenefitSystem
from openfisca_us.entities import *
from openfisca_us.situation_examples import single_filer
import os

COUNTRY_DIR = os.path.dirname(os.path.abspath(__file__))


# Our country tax and benefit class inherits from the general TaxBenefitSystem class.
# The name CountryTaxBenefitSystem must not be changed, as all tools of the OpenFisca ecosystem expect a CountryTaxBenefitSystem class to be exposed in the __init__ module of a country package.
class CountryTaxBenefitSystem(TaxBenefitSystem):
    CURRENCY = "$"

    def __init__(self):
        # We initialize our tax and benefit system with the general constructor
        super().__init__(entities)

        # We add to our tax and benefit system all the variables
        self.add_variables_from_directory(
            os.path.join(COUNTRY_DIR, "variables")
        )

        # We add to our tax and benefit system all the legislation parameters defined in the  parameters files
        param_path = os.path.join(COUNTRY_DIR, "parameters")
        self.load_parameters(param_path)

        # We define which variable, parameter and simulation example will be used in the OpenAPI specification



from openfisca_us.model_api import *


class has_elderly_disabled(Variable):
    value_type = bool
    entity = SPMUnit
    documentation = (
        "Whether elderly or disabled people, per USDA definitions, are present"
    )
    label = "Elderly or disabled person present"
    definition_period = YEAR



from openfisca_core.model_api import *
from openfisca_us.entities import *


class ProviderTypeGroup(Enum):
    DCC_SACC = "Licenced/registered/permitted day care center; registered school-age child care"
    FDC_GFDC = (
        "Registered family day care homes; licensed group family day care"
    )
    LE_GC = "Legally exempt group child care programs"
    LE_STD = "Informal child care standard rate"
    LE_ENH = "Informal child care enhanced rate"


class provider_type_group(Variable):
    value_type = Enum
    possible_values = ProviderTypeGroup
    # DCC_SACC is most common among provider types
    default_value = ProviderTypeGroup.DCC_SACC
    entity = Person
    label = u"CCDF provider type group"
    definition_period = YEAR


class childcare_hours_per_week(Variable):
    value_type = float
    entity = Person
    label = u"Child care hours per week"
    definition_period = YEAR


class childcare_hours_per_day(Variable):
    value_type = float
    entity = Person
    label = u"Child care hours per day"
    definition_period = YEAR


class childcare_days_per_week(Variable):
    value_type = float
    entity = Person
    label = u"Child care days per week"
    definition_period = YEAR



from openfisca_us.model_api import *


class snap_net_income(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = YEAR
    documentation = "Final net income, after all deductions"
    label = "SNAP net income"
    unit = "currency-USD"
    reference = "https://www.law.cornell.edu/uscode/text/7/2014"

    def formula(spm_unit, period):
        gross_income = spm_unit("snap_gross_income", period)
        deductions = spm_unit("snap_deductions", period)
        return max_(0, gross_income - deductions)



from openfisca_tools.model_api import *
from openfisca_us.entities import *


class household_net_income(Variable):
    value_type = float
    entity = Household
    label = u"Household net income"
    definition_period = YEAR

    def formula(household, period, parameters):
        return household.sum(household.members("net_income", period))
